import uuid
from collections import defaultdict
from slugify import slugify

from django.core.exceptions import FieldError, ValidationError
from django.db import models
from django.db.models.query import ModelIterable
from django.utils.functional import cached_property
from django.utils.translation import gettext as _

from arches import VERSION as arches_version
from arches.app.models.models import Node

from arches_querysets.datatypes.datatypes import DataTypeFactory
from arches_querysets.utils.models import (
    generate_node_alias_expressions,
    get_recursive_prefetches,
)

NOT_PROVIDED = object()


class NodeAliasValuesMixin:
    def values(self, *args, **kwargs):
        """Allow using a node_alias as a field name in a values() query."""
        # This is just sugar so that the following works:
        # .values("my_alias")
        # rather than the long form:
        # .values(my_alias=models.F("my_alias"))
        args_copy = list(set(args))
        kwargs_copy = {**kwargs}
        for arg in args:
            field_name = arg.split(models.constants.LOOKUP_SEP)[0]
            if (
                field_name in self.query.annotations
                and field_name not in self.query.annotation_select
                and arg not in kwargs
            ):
                # values() can promote aliases to annotations via **kwargs
                kwargs_copy[arg] = models.F(arg)
                args_copy.remove(arg)
        return super().values(*args_copy, **kwargs_copy)

    def values_list(self, *args, **kwargs):
        """Allow using a node_alias as a field name in a values_list() query."""
        qs = self
        for arg in args:
            field_name = arg.split(models.constants.LOOKUP_SEP)[0]
            if (
                field_name in qs.query.annotations
                and field_name not in qs.query.annotation_select
            ):
                # values_list() cannot promote aliases via kwargs like annotate().
                qs = qs.annotate(**{arg: models.F(arg)})
        return models.QuerySet.values_list(qs, *args, **kwargs)

    def aggregate(self, *args, **kwargs):
        """Handle the "promotion" of .alias() to .annotate() for .aggregate()."""
        try:
            return super().aggregate(*args, **kwargs)
        except FieldError as e:
            # Hopefully Django will support this out of the box, see:
            # https://code.djangoproject.com/ticket/36480#comment:3
            node_alias = e.args[0].split("'")[1]
            with_annotation = self.annotate(**{node_alias: models.F(node_alias)})
            return with_annotation.aggregate(*args, **kwargs)


class TileTreeManager(models.Manager):
    # TODO: evaluate whether some of these should get pushed into get_tiles()
    def get_queryset(self):
        qs = super().get_queryset().select_related("resourceinstance")
        # arches_version==9.0.0
        if arches_version >= (8, 0):
            # TODO: could get this once?
            qs = qs.prefetch_related("resourceinstance__from_resxres__to_resource")
            qs = qs.select_related("nodegroup__grouping_node")
        else:
            # Annotate nodegroup_alias on Arches 7.6.
            qs = qs.annotate(
                _nodegroup_alias=Node.objects.filter(
                    pk=models.F("nodegroup_id"),
                    nodegroup__tilemodel=models.OuterRef("tileid"),
                ).values("alias")[:1]
            )
            qs = qs.prefetch_related(
                "nodegroup__nodegroup_set__node_set",
                "resourceinstance__resxres_resource_instance_ids_from__resourceinstanceidto",
            )
        return qs


class TileTreeQuerySet(NodeAliasValuesMixin, models.QuerySet):
    def get_tiles(
        self,
        graph_slug,
        nodegroup_alias=None,
        *,
        resource_ids=None,
        as_representation=False,
        depth=20,
        nodes=None,
        graph_query=None,
    ):
        """
        Entry point for filtering arches data by nodegroups.

        >>> statements = TileTree.get_tiles("datatype_lookups", "statement")
        # TODO: show this with some test node that's actually a localized string.
        >>> results = statements.filter(statement_content__any_lang_startswith="F")
        >>> for result in results:
                print(result.resourceinstance)
                print("\t", result.statement_content["en"]["value"])  # TODO: unwrap?

        <Concept: x-ray fluorescence (aec56d59-9292-42d6-b18e-1dd260ff446f)>
            Fluorescence stimulated by x-rays; ...
        <Concept: vellum (parchment) (34b081cd-6fcc-4e00-9a43-0a8a73745b45)>
            Fine-quality calf or lamb parchment ...

        as_representation:
            - True: calls to_json() datatype methods
            - False: calls to_python() datatype methods
        """
        from arches_querysets.models import GraphWithPrefetching

        if nodegroup_alias:
            qs = self.filter(
                nodegroup__node__alias=nodegroup_alias,
                nodegroup__node__graph__slug=graph_slug,
            )
        else:
            qs = self.filter(nodegroup__node__graph__slug=graph_slug)
        if resource_ids:
            qs = qs.filter(resourceinstance_id__in=resource_ids)

        if not nodes:
            # Violates laziness of QuerySets, but can be made fully lazy
            # by providing a `nodes` argument doing the same query.
            filters = models.Q(graph__slug=graph_slug)
            # arches_version==9.0.0
            if arches_version >= (8, 0):
                filters &= models.Q(source_identifier=None)
            nodes = (
                Node.objects.filter(filters)
                .exclude(datatype="semantic")
                .exclude(nodegroup=None)
                .select_related("nodegroup__parentnodegroup")
            )
            if not nodes:
                raise ValueError(f"No nodes found for graph with slug: {graph_slug}")

        alias_expressions = generate_node_alias_expressions(self.model, nodes)

        # arches_version==9.0.0
        if arches_version < (8, 0):
            msg = "arches-querysets requires all nodes to have an alias."
            assert None not in alias_expressions, msg

        if graph_query is None:
            if "graph_query" in qs._hints:
                graph_query = qs._hints["graph_query"]
            else:
                graph_query = GraphWithPrefetching.objects.prefetch(graph_slug)

        qs._add_hints(
            as_representation=as_representation,
            graph_slug=graph_slug,
            graph_query=graph_query,
        )

        # Future: see various solutions mentioned here for avoiding
        # "magic number" depth traversal (but the magic number is harmless,
        # causes no additional queries beyond actual depth):
        # https://forum.djangoproject.com/t/prefetching-relations-to-arbitrary-depth/39328
        if depth:
            child_tile_query = self.model.objects.get_queryset().get_tiles(
                graph_slug=graph_slug,
                as_representation=as_representation,
                depth=depth - 1,
                nodes=nodes,
                graph_query=graph_query,
            )

            qs = qs.prefetch_related(
                models.Prefetch(
                    # arches_version==9.0.0
                    "children" if arches_version >= (8, 0) else "tilemodel_set",
                    queryset=child_tile_query,
                    # Using to_attr ensures the query results materialize into
                    # TileTree objects rather than TileModel objects. This isn't
                    # usually an issue, but something in the way we're overriding
                    # ORM internals seems to require this.
                    to_attr="_tile_trees",
                )
            )

        # Provide the alias_expressions to the ORM.
        # Use .distinct() because the inner join above ("nodegroup__node__graph__slug")
        # can produce duplicates.
        return qs.alias(**alias_expressions).order_by("sortorder").distinct()

    @cached_property
    def grouping_node_lookup(self):
        graph_nodes = next(iter(self._hints["graph_query"])).node_set.all()
        return {node.pk: node for node in graph_nodes if node.pk == node.nodegroup_id}

    def _fetch_all(self):
        """Hook into QuerySet evaluation to customize the result."""
        if issubclass(self._iterable_class, ModelIterable):
            if "graph_query" in self._hints:
                self._iterable_class = TileTreeIterable
            # else: .get_tiles() was not called: no need to set richer representations.
        # else: values()/values_list() queries: no need to set richer representations.
        super()._fetch_all()

        if issubclass(self._iterable_class, ModelIterable):
            try:
                self._set_aliased_data()
            except (TypeError, ValueError, ValidationError) as e:
                # These errors are caught by DRF, so re-raise as something else.
                raise RuntimeError(e) from e

    def _set_aliased_data(self):
        """
        Call datatype to_python() methods when materializing the QuerySet.
        Memoize fetched nodes.
        Fetch display values in bulk.
        Attach child tiles to parent tiles and vice versa.
        """
        from arches_querysets.models import AliasedData

        aliased_data_to_update = {}
        values_by_datatype = defaultdict(list)
        datatype_contexts = {}
        for tile in self._result_cache:
            if tile.aliased_data is None:
                tile.aliased_data = AliasedData()
            else:
                return  # already set
            tile.sync_private_attributes(self)
            for node in tile.nodegroup.node_set.all():
                if node.datatype == "semantic":
                    continue
                datatype_instance = DataTypeFactory().get_instance(node.datatype)
                tile_data = datatype_instance.get_tile_data(tile)
                node_value = tile_data.get(str(node.pk))
                if node_value is None:
                    # Datatype methods assume tiles always have all keys, but we've
                    # seen problems in the wild.
                    tile_data[str(node.pk)] = None
                aliased_data_to_update[(tile, node)] = node_value
                values_by_datatype[node.datatype].append(node_value)

        # Get datatype context querysets.
        for datatype, values in values_by_datatype.items():
            datatype_instance = DataTypeFactory().get_instance(datatype)
            bulk_values = datatype_instance.get_display_value_context_in_bulk(values)
            datatype_instance.set_display_value_context_in_bulk(bulk_values)
            datatype_contexts[datatype] = bulk_values

        # Set aliased_data property.
        for tile_node_pair, node_value in aliased_data_to_update.items():
            tile, node = tile_node_pair
            tile.set_aliased_data(node, node_value, datatype_contexts)

        for tile in self._result_cache:
            self._set_child_tile_data(tile)

    def _set_child_tile_data(self, tile):
        child_tiles = getattr(tile, "_tile_trees", [])
        for child_tile in sorted(
            child_tiles, key=lambda tile_item: tile_item.sortorder or 0
        ):
            child_nodegroup_alias = child_tile.find_nodegroup_alias(
                self.grouping_node_lookup
            )
            if child_tile.nodegroup.cardinality == "1" and child_nodegroup_alias:
                # TODO(arches_version==9.0.0): remove `and child_nodegroup_alias`
                # which can no longer be null as of v8.
                setattr(tile.aliased_data, child_nodegroup_alias, child_tile)
            else:
                children = getattr(tile.aliased_data, child_nodegroup_alias, [])
                children.append(child_tile)
                setattr(tile.aliased_data, child_nodegroup_alias, children)
            # Attach parent to this child.
            child_tile.parent = tile
            child_tile.sync_private_attributes(tile)

        child_nodegroups = (
            getattr(tile.nodegroup, "children")
            # arches_version==9.0.0
            if arches_version >= (8, 0)
            else getattr(tile.nodegroup, "nodegroup_set")
        )
        for child_nodegroup in child_nodegroups.all():
            for node in child_nodegroup.node_set.all():
                if node.pk == child_nodegroup.pk:
                    grouping_node = node
                    break

            if (
                getattr(tile.aliased_data, grouping_node.alias, NOT_PROVIDED)
                is NOT_PROVIDED
            ):
                setattr(
                    tile.aliased_data,
                    grouping_node.alias,
                    None if child_nodegroup.cardinality == "1" else [],
                )


class TileTreeIterable(ModelIterable):
    def __iter__(self):
        """
        Ensure every tile and nodegroup in the tree has a reference to the
        prefetched graph via
        - tile -> resourceinstance -> graph
        - tile -> nodegroup -> grouping_node & node_set

        Set .aliased_data to None as a sentinel so that TileTreeQuerySet._set_aliased_data()
        knows to run only once.
        """
        graph = next(iter(self.queryset._hints["graph_query"]))
        nodegroup_map = {
            node.nodegroup_id: node.nodegroup for node in graph.node_set.all()
        }
        for tile_tree in super().__iter__():
            if not tile_tree.sealed:
                tile_tree.resourceinstance.graph = graph
                tile_tree.nodegroup = nodegroup_map[tile_tree.nodegroup_id]
                tile_tree.sealed = True
                tile_tree.aliased_data = None
            yield tile_tree


class ResourceTileTreeQuerySet(NodeAliasValuesMixin, models.QuerySet):
    def get_tiles(
        self,
        graph_slug,
        *,
        resource_ids=None,
        as_representation=False,
        nodes=None,
        depth=20,
    ):
        """Aliases a ResourceTileTreeQuerySet with tile data unpacked
        and mapped onto nodegroup aliases, e.g.:

        >>> concepts = ResourceTileTree.objects.get_tiles("concept")

        With slightly fewer keystrokes:

        >>> concepts = ResourceTileTree.get_tiles("concept")

        Django QuerySet methods are available for efficient queries:
        >>> concepts.count()
        785

        Filter on any nested node at the top level ("shallow query").
        In this example, statement_content is a cardinality-N node, thus an array.

        >>> subset = concepts.filter(statement_content__len__gt=0)[:4]
        >>> for concept in subset:
                print(concept)
                for stmt in concept.aliased_data.statement:
                    print("\t", stmt)
                    print("\t\t", stmt.aliased_data.statement_content)

        <Concept: consignment (method of acquisition) (f3fed7aa-eae6-41f6-aa0f-b889d84c0552)>
            <TileModel: statement (46efcd06-a5e5-43be-8847-d7cd94cbc9cb)>
                'Individual objects or works. Most works ...
        ...

        Access child and parent tiles by nodegroup aliases:

        # TODO: replace this example.
        >>> has_child = concepts.filter(statement_data_assignment_statement_content__len__gt=0).first()
        >>> has_child
        <Concept: <appellative_status_ascribed_name_content> (751614c0-de7a-47d7-8e87-a4d18c7337ff)>
        >>> has_child.statement_data_assignment_statement
        <statement_data_assignment_statement (51e1f473-712e-447b-858e-cc7353a084a6)>
        >>> parent = has_child.statement[0]
        >>> parent.statement_data_assignment_statement[0].statement is parent
        True

        Provisional edits are completely ignored for the purposes of querying.

        as_representation:
            - True: calls to_json() datatype methods
            - False: calls to_python() datatype methods
        """
        from arches_querysets.models import GraphWithPrefetching, TileTree

        graph_query = GraphWithPrefetching.objects.prefetch(graph_slug)
        self._add_hints(as_representation=as_representation, graph_query=graph_query)

        if not nodes:
            # Violates laziness of QuerySets, but can be made fully lazy
            # by providing a `nodes` argument doing the same query.
            filters = models.Q(graph__slug=graph_slug)
            # arches_version==9.0.0
            if arches_version >= (8, 0):
                filters &= models.Q(source_identifier=None)
            nodes = (
                Node.objects.filter(filters)
                .exclude(datatype="semantic")
                .exclude(nodegroup=None)
                .select_related("nodegroup__parentnodegroup")
                .order_by()
            )
            if not nodes:
                raise ValueError(f"No nodes found for graph with slug: {graph_slug}")

        alias_expressions = generate_node_alias_expressions(self.model, nodes)

        if resource_ids:
            qs = self.filter(graph__slug=graph_slug, pk__in=resource_ids)
        else:
            filters = models.Q(graph__slug=graph_slug)
            # arches_version==9.0.0
            if arches_version >= (8, 0):
                filters &= models.Q(graph__source_identifier=None)
            qs = self.filter(filters)

        return (
            qs.select_related("graph")
            .prefetch_related(
                models.Prefetch(
                    "tilemodel_set",
                    queryset=TileTree.objects.get_tiles(
                        graph_slug=graph_slug,
                        as_representation=as_representation,
                        nodes=nodes,
                        graph_query=graph_query,
                        depth=depth,
                    ),
                    to_attr="_tile_trees",
                ),
            )
            .alias(**alias_expressions)
        )

    def _fetch_all(self):
        """Hook into QuerySet evaluation to customize the result."""
        if issubclass(self._iterable_class, ModelIterable):
            if "graph_query" in self._hints:
                self._iterable_class = ResourceTileTreeIterable
            # else: .get_tiles() was not called: no need to set richer representations.
        # else: values()/values_list() queries: no need to set richer representations.

        super()._fetch_all()

        if issubclass(self._iterable_class, ModelIterable):
            try:
                self._set_aliased_data()
            except (TypeError, ValueError, ValidationError) as e:
                # These errors are caught by DRF, so re-raise as something else.
                raise RuntimeError(e) from e
        # else: values()/values_list() queries

    def _set_aliased_data(self):
        """
        Attach top-level tiles to resource instances.
        Attach resource instances to all fetched tiles.
        Memoize fetched grouping node aliases (and graph source nodes).
        """
        from arches_querysets.models import AliasedData, GraphWithPrefetching

        if not self._result_cache:
            return
        for resource in self._result_cache:
            if resource.aliased_data is None:
                resource.aliased_data = AliasedData()
            else:
                return  # already processed
        for resource in self._result_cache:
            for tile in getattr(resource, "_tile_trees", []):
                graph = tile.resourceinstance.graph
                break
            else:
                continue
            break
        else:
            graph = GraphWithPrefetching.objects.prefetch(
                graph_slug=self._result_cache[0].graph.slug
            ).get()

        grouping_nodes = {}
        for node in graph.node_set.all():
            if not node.nodegroup:
                continue
            if getattr(node, "source_identifier", None):
                continue
            if node.pk == node.nodegroup_id:
                grouping_nodes[node.pk] = node

        for resource in self._result_cache:
            resource._as_representation = self._hints.get("as_representation", False)

            # Prepare empty aliased data containers.
            for grouping_node in grouping_nodes.values():
                if grouping_node.nodegroup.parentnodegroup_id:
                    continue
                default = None if grouping_node.nodegroup.cardinality == "1" else []
                setattr(resource.aliased_data, grouping_node.alias, default)

            # Fill aliased data with top nodegroup data.
            for tile in getattr(resource, "_tile_trees", []):
                if tile.nodegroup.parentnodegroup_id:
                    continue
                nodegroup_alias = grouping_nodes[tile.nodegroup_id].alias
                if tile.nodegroup.cardinality == "n":
                    tile_array = getattr(resource.aliased_data, nodegroup_alias)
                    tile_array.append(tile)
                else:
                    setattr(resource.aliased_data, nodegroup_alias, tile)


class ResourceTileTreeIterable(ModelIterable):
    def __iter__(self):
        """
        Set .aliased_data to None as a sentinel so that
        ResourceTileTreeQuerySet._set_aliased_data() knows to run only once.
        """
        graph = next(iter(self.queryset._hints["graph_query"]))
        for resource_tile_tree in super().__iter__():
            if not resource_tile_tree.sealed:
                resource_tile_tree.graph = graph
                resource_tile_tree.sealed = True
                resource_tile_tree.aliased_data = None
            yield resource_tile_tree


# TODO (arches_version==9.0.0): remove all but prefetch() method dropping 7.6
class GraphWithPrefetchingQuerySet(models.QuerySet):  # pragma: no cover
    """Backport of Arches 8.0 GraphQuerySet, plus one method for custom prefetches."""

    def make_name_unique(self, name, names_to_check, suffix_delimiter="_"):
        """
        Makes a name unique among a list of names

        Arguments:
        name -- the name to check and modfiy to make unique in the list of "names_to_check"
        names_to_check -- a list of names that "name" should be unique among
        """

        i = 1
        temp_node_name = name
        while temp_node_name in names_to_check:
            temp_node_name = "{0}{1}{2}".format(name, suffix_delimiter, i)
            i += 1
        return temp_node_name

    def create(self, *args, **kwargs):
        raise NotImplementedError(
            "Use create_graph() to create new Graph instances with proper business logic."
        )

    def generate_slug(self, name, is_resource):
        if name:
            slug = slugify(name, separator="_")
        else:
            if is_resource:
                slug = "new_resource_model"
            else:
                slug = "new_branch"
        existing_slugs = self.values_list("slug", flat=True)
        slug = self.make_name_unique(slug, existing_slugs, "_")

        return slug

    def create_graph(self, name="", *, slug=None, user=None, is_resource=False):
        from arches.app.models import models as arches_models
        from arches.app.models.graph import Graph as OldGraphWithPrefetchingModel

        """
        Create a new Graph and related objects, encapsulating all creation side effects.
        """
        new_id = uuid.uuid4()
        nodegroup = None

        if not slug:
            slug = self.generate_slug(name, is_resource)

        graph_model = arches_models.GraphModel(
            name=name,
            subtitle="",
            author=(
                " ".join(filter(None, [user.first_name, user.last_name]))
                if user
                else ""
            ),
            description="",
            version="",
            isresource=is_resource,
            iconclass="",
            ontology=None,
            slug=slug,
        )
        graph_model.save()  # to access side-effects declared in save method

        if not is_resource:
            nodegroup = arches_models.NodeGroup.objects.create(pk=new_id)
            arches_models.CardModel.objects.create(
                nodegroup=nodegroup, name=name, graph=graph_model
            )

        # root node
        arches_models.Node.objects.create(
            pk=new_id,
            name=name,
            description="",
            istopnode=True,
            ontologyclass=None,
            datatype="semantic",
            nodegroup=nodegroup,
            graph=graph_model,
        )

        graph = OldGraphWithPrefetchingModel.objects.get(pk=graph_model.graphid)

        graph.publish(
            user=user,
            notes=_("Graph created."),
        )
        if arches_version >= (8, 0):
            graph.create_draft_graph()

        # ensures entity returned matches database entity
        return self.get(pk=graph_model.graphid)

    def prefetch(self, graph_slug=None, *, resource_ids=None):
        """Return a graph with necessary prefetches for setting aliased_data."""
        qs = self
        if resource_ids and not graph_slug:
            qs = qs.filter(resourceinstance__in=resource_ids)
        elif graph_slug:
            # arches_version==9.0.0
            if arches_version >= (8, 0):
                qs = qs.filter(slug=graph_slug, source_identifier=None)
            else:
                qs = qs.filter(slug=graph_slug)
        else:
            raise ValueError("graph_slug or resource_ids must be provided")

        # arches_version==9.0.0
        if arches_version >= (8, 0):
            children = "children"
        else:
            children = "nodegroup_set"

        prefetches = [
            "node_set__cardxnodexwidget_set",
            "node_set__nodegroup__parentnodegroup",
            "node_set__nodegroup__node_set__cardxnodexwidget_set",
            "node_set__nodegroup__cardmodel_set",
            *get_recursive_prefetches(
                f"node_set__nodegroup__{children}", depth=12, recursive_part=children
            ),
            *get_recursive_prefetches(
                f"node_set__nodegroup__{children}__node_set__cardxnodexwidget_set",
                depth=12,
                recursive_part=children,
            ),
            *get_recursive_prefetches(
                f"node_set__nodegroup__{children}__cardmodel_set",
                depth=12,
                recursive_part=children,
            ),
        ]

        return qs.prefetch_related(*prefetches)[:1]

    def _fetch_all(self):
        """Derive each grouping_node instead of fetching it."""
        super()._fetch_all()
        if not isinstance(self._iterable_class, ModelIterable):
            return
        for graph in self._result_cache or []:  # XXX consider memo'ing this as done.
            for node in graph.node_set.all():
                if node.nodegroup:
                    for sibling_node in node.nodegroup.node_set.all():
                        if sibling_node.pk == node.nodegroup.pk:
                            node.nodegroup.grouping_node = sibling_node
                            break



import unittest
import uuid
from http import HTTPStatus
from unittest.mock import patch

from django.core.management import call_command
from django.urls import reverse
from arches import VERSION as arches_version
from arches.app.models.graph import Graph
from arches.app.models.models import EditLog

from arches_querysets.rest_framework.serializers import (
    ArchesResourceSerializer,
    ArchesResourceTopNodegroupsSerializer,
    ArchesSingleNodegroupSerializer,
    ArchesTileSerializer,
)
from arches_querysets.utils.models import ensure_request
from arches_querysets.utils.tests import GraphTestCase


MUTABLE_PERMITTED_NODEGROUPS = set()


class RestFrameworkTests(GraphTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        call_command("add_test_users", verbosity=0)

    def patched_ensure_request(self, request, force_admin):
        request.user.userprofile.viewable_nodegroups = {str(self.nodegroup_id)}
        self.set_single_viewable_nodegroup(request, self.nodegroup_1.pk)
        return ensure_request(request, force_admin)

    def test_create_tile_for_new_resource(self):
        create_url = reverse(
            "arches_querysets:api-tiles",
            kwargs={"graph": "datatype_lookups", "nodegroup_alias": "datatypes_n"},
        )
        request_body = {"aliased_data": {"string_alias_n": "create_value"}}

        # Anonymous user lacks editing permissions.
        forbidden_response = self.client.post(
            create_url, request_body, content_type="application/json"
        )
        self.assertEqual(forbidden_response.status_code, HTTPStatus.FORBIDDEN)

        # Dev user can edit.
        self.client.login(username="dev", password="dev")
        response = self.client.post(
            create_url, request_body, content_type="application/json"
        )

        # The response includes the context.
        self.assertEqual(response.status_code, HTTPStatus.CREATED, response.content)
        self.assertIn("aliased_data", response.json())
        self.assertIsInstance(uuid.UUID(response.json()["tileid"]), uuid.UUID)
        self.assertEqual(
            response.json()["aliased_data"]["string_alias_n"],
            {
                "display_value": "create_value",
                "node_value": {
                    "en": {"value": "create_value", "direction": "ltr"},
                },
                "details": [],
            },
        )
        self.assertSequenceEqual(
            EditLog.objects.filter(
                resourceinstanceid=response.json()["resourceinstance"],
            )
            .values_list("edittype", flat=True)
            .order_by("edittype"),
            ["create", "tile create"],
        )

    def test_create_tile_for_existing_resource(self):
        create_url = reverse(
            "arches_querysets:api-tiles",
            kwargs={"graph": "datatype_lookups", "nodegroup_alias": "datatypes_n"},
        )
        request_body = {
            "aliased_data": {"string_alias_n": "create_value"},
            "resourceinstance": str(self.resource_42.pk),
        }
        self.client.login(username="dev", password="dev")
        response = self.client.post(
            create_url, request_body, content_type="application/json"
        )
        self.assertEqual(response.status_code, HTTPStatus.CREATED)
        self.assertIsInstance(uuid.UUID(response.json()["tileid"]), uuid.UUID)
        self.assertEqual(response.json()["resourceinstance"], str(self.resource_42.pk))
        self.assertEqual(
            response.json()["aliased_data"]["string_alias_n"],
            {
                "display_value": "create_value",
                "node_value": {
                    "en": {"value": "create_value", "direction": "ltr"},
                },
                "details": [],
            },
        )

    def test_create_nested_tiles_for_new_resource(self):
        self.client.login(username="dev", password="dev")
        create_url = reverse(
            "arches_querysets:api-tiles",
            kwargs={"graph": "datatype_lookups", "nodegroup_alias": "datatypes_1"},
        )
        request_body = {
            "aliased_data": {
                "string_alias": "create_value",
                "datatypes_1_child": {
                    "aliased_data": {"string_alias_child": "child_create_value"}
                },
            },
        }

        response = self.client.post(
            create_url, request_body, content_type="application/json"
        )

        # The response includes the context.
        self.assertEqual(response.status_code, HTTPStatus.CREATED)
        self.assertIn("aliased_data", response.json())
        self.assertIsInstance(uuid.UUID(response.json()["tileid"]), uuid.UUID)
        parent_data = response.json()["aliased_data"]
        self.assertEqual(
            parent_data["datatypes_1_child"]["aliased_data"]["string_alias_child"],
            {
                "display_value": "child_create_value",
                "node_value": {
                    "en": {"value": "child_create_value", "direction": "ltr"},
                },
                "details": [],
            },
        )
        self.assertEqual(response.status_code, HTTPStatus.CREATED, response.content)

        self.assertSequenceEqual(
            EditLog.objects.filter(
                resourceinstanceid=response.json()["resourceinstance"],
            )
            .values_list("edittype", flat=True)
            .order_by("edittype"),
            ["create", "tile create", "tile create"],
        )

    def test_create_nested_tiles_for_new_resource_via_resource_serializer(self):
        self.client.login(username="dev", password="dev")
        create_url = reverse(
            "arches_querysets:api-resources",
            kwargs={"graph": "datatype_lookups"},
        )
        request_body = {
            "aliased_data": {
                "datatypes_1": {
                    "aliased_data": {
                        "string_alias": "create_value",
                        "datatypes_1_child": {
                            "aliased_data": {"string_alias_child": "child_create_value"}
                        },
                    },
                },
            },
        }

        response = self.client.post(
            create_url, request_body, content_type="application/json"
        )

        # The response includes the context.
        self.assertEqual(response.status_code, HTTPStatus.CREATED, response.json())
        parent_data = response.json()["aliased_data"]["datatypes_1"]["aliased_data"]
        self.assertEqual(
            parent_data["datatypes_1_child"]["aliased_data"]["string_alias_child"],
            {
                "display_value": "child_create_value",
                "node_value": {
                    "en": {"value": "child_create_value", "direction": "ltr"},
                },
                "details": [],
            },
        )
        self.assertEqual(response.status_code, HTTPStatus.CREATED, response.content)

        edits = EditLog.objects.filter(
            resourceinstanceid=response.json()["resourceinstanceid"],
        ).order_by("edittype")
        self.assertEqual(
            [edit.edittype for edit in edits], ["create", "tile create", "tile create"]
        )
        self.assertEqual(len(set([edit.transactionid for edit in edits])), 1)

    def test_update_tile(self):
        update_url = reverse(
            "arches_querysets:api-tile",
            kwargs={
                "graph": "datatype_lookups",
                "nodegroup_alias": "datatypes_1",
                "pk": self.resource_42.aliased_data.datatypes_1.pk,
            },
        )
        request_body = {
            "aliased_data": {"string_alias": "update_value"},
            "resourceinstance": str(self.resource_42.pk),
        }
        assert "tileid" not in request_body, "tileid is not required in update requests"

        self.client.login(username="dev", password="dev")
        response = self.client.patch(
            update_url, request_body, content_type="application/json"
        )
        self.assertEqual(response.status_code, HTTPStatus.OK, response.json())
        self.assertIsInstance(uuid.UUID(response.json()["tileid"]), uuid.UUID)
        self.assertEqual(response.json()["resourceinstance"], str(self.resource_42.pk))
        self.assertEqual(
            response.json()["aliased_data"]["string_alias"],
            {
                "display_value": "update_value",
                "node_value": {
                    "en": {"value": "update_value", "direction": "ltr"},
                },
                "details": [],
            },
        )

    @unittest.skipIf(arches_version < (8, 0), reason="Arches 8+ only logic")
    def test_out_of_date_resource(self):
        Graph.objects.get(pk=self.graph.pk).publish(user=None)

        update_url = reverse(
            "arches_querysets:api-resource",
            kwargs={"graph": "datatype_lookups", "pk": str(self.resource_42.pk)},
        )
        self.client.login(username="dev", password="dev")
        request_body = {"aliased_data": {"datatypes_1": None}}
        response = self.client.put(
            update_url, request_body, content_type="application/json"
        )
        self.assertContains(
            response,
            "Graph Has Different Publication",
            status_code=HTTPStatus.BAD_REQUEST,
        )

    def test_instantiate_empty_resource_serializer(self):
        serializer = ArchesResourceSerializer(graph_slug="datatype_lookups")
        self.assertIsNone(serializer.data["resourceinstanceid"])
        # Default values are stocked.
        self.assertEqual(
            serializer.data["aliased_data"]["datatypes_1"]["aliased_data"][
                "number_alias"
            ]["node_value"],
            7,
        )

    def test_instantiate_empty_tile_serializer(self):
        serializer = ArchesTileSerializer(
            graph_slug="datatype_lookups", nodegroup_alias="datatypes_1"
        )
        self.assertIsNone(serializer.data["tileid"])
        # Default values are stocked.
        self.assertEqual(
            serializer.data["aliased_data"]["number_alias"]["node_value"], 7
        )

    def test_bind_data_to_serializer(self):
        # Get some default data from the serializer.
        static_data = ArchesTileSerializer(
            graph_slug="datatype_lookups", nodegroup_alias="datatypes_1"
        ).data
        # Pretend that data came from somewhere else, and process it, e.g. in a script.
        serializer = ArchesTileSerializer(
            graph_slug="datatype_lookups",
            nodegroup_alias="datatypes_1",
            data=static_data,
        )
        self.assertTrue(serializer.is_valid())

        # Or, submit it to the API
        self.client.login(username="dev", password="dev")
        create_url = reverse(
            "arches_querysets:api-tiles",
            kwargs={"graph": "datatype_lookups", "nodegroup_alias": "datatypes_1"},
        )
        response = self.client.post(
            create_url, serializer.data, content_type="application/json"
        )
        self.assertEqual(response.status_code, HTTPStatus.CREATED)

    def test_exclude_children_option(self):
        serializer = ArchesResourceSerializer(graph_slug="datatype_lookups")
        self.assertIn(
            "datatypes_1_child",
            serializer.data["aliased_data"]["datatypes_1"]["aliased_data"],
        )
        serializer = ArchesResourceTopNodegroupsSerializer(
            graph_slug="datatype_lookups"
        )
        self.assertNotIn(
            "datatypes_1_child",
            serializer.data["aliased_data"]["datatypes_1"]["aliased_data"],
        )
        serializer = ArchesTileSerializer(
            graph_slug="datatype_lookups", nodegroup_alias="datatypes_1"
        )
        self.assertIn("datatypes_1_child", serializer.data["aliased_data"])
        serializer = ArchesSingleNodegroupSerializer(
            graph_slug="datatype_lookups", nodegroup_alias="datatypes_1"
        )
        self.assertNotIn("datatypes_1_child", serializer.data["aliased_data"])

    def test_blank_views_tile_id(self):
        response = self.client.get(
            reverse(
                "arches_querysets:api-tile-blank",
                kwargs={"graph": "datatype_lookups", "nodegroup_alias": "datatypes_1"},
            )
        )
        self.assertIsNone(response.json()["tileid"])

    def test_blank_views_exclude_children_option(self):
        response = self.client.get(
            reverse(
                "arches_querysets:api-resource-blank",
                kwargs={"graph": "datatype_lookups"},
            )
        )
        self.assertContains(response, "datatypes_1_child")

        response = self.client.get(
            reverse(
                "arches_querysets:api-resource-blank",
                kwargs={"graph": "datatype_lookups"},
            ),
            QUERY_STRING="exclude_children=true",
        )
        self.assertNotContains(response, "datatypes_1_child")

        response = self.client.get(
            reverse(
                "arches_querysets:api-tile-blank",
                kwargs={"graph": "datatype_lookups", "nodegroup_alias": "datatypes_1"},
            )
        )
        self.assertContains(response, "datatypes_1_child")

        response = self.client.get(
            reverse(
                "arches_querysets:api-tile-blank",
                kwargs={
                    "graph": "datatype_lookups",
                    "nodegroup_alias": "datatypes_1",
                },
            ),
            QUERY_STRING="exclude_children=true",
        )
        self.assertNotContains(response, "datatypes_1_child")

    def test_fill_blanks_option(self):
        self.resource_42.tilemodel_set.all().delete()
        response = self.client.get(
            reverse(
                "arches_querysets:api-resource",
                kwargs={"graph": "datatype_lookups", "pk": str(self.resource_42.pk)},
            ),
            QUERY_STRING="fill_blanks=true",
        )
        parent_data = response.json()["aliased_data"]["datatypes_1"]
        self.assertIsNone(parent_data["tileid"])
        child_data = parent_data["aliased_data"]["datatypes_1_child"]
        self.assertIsNone(child_data["tileid"])

    @patch(
        "arches.app.models.models.UserProfile.viewable_nodegroups",
        MUTABLE_PERMITTED_NODEGROUPS,
    )
    def test_serializer_observes_nodegroup_permissions(self):
        resource_serializer = ArchesResourceSerializer(graph_slug="datatype_lookups")
        self.assertNotIn("datatypes_1", resource_serializer.data["aliased_data"])

        # A TileSerializer where the topmost nodegroup is not permitted raises
        tile_serializer = ArchesTileSerializer(
            graph_slug="datatype_lookups", nodegroup_alias="datatypes_1"
        )
        with self.assertRaises(PermissionError):
            tile_serializer.data

        # Otherwise we just return whatever part of the tree we can.
        MUTABLE_PERMITTED_NODEGROUPS.add(str(self.nodegroup_1.pk))
        tile_serializer = ArchesTileSerializer(
            graph_slug="datatype_lookups", nodegroup_alias="datatypes_1"
        )
        self.assertIn("number_alias", tile_serializer.data["aliased_data"])
        self.assertNotIn("datatypes_1_child", tile_serializer.data["aliased_data"])

    def test_filter_kwargs(self):
        node_alias = "string_alias"

        response = self.client.get(
            reverse(
                "arches_querysets:api-resources",
                kwargs={"graph": "datatype_lookups"},
            ),
            # Additional lookups tested in test_lookups.py
            QUERY_STRING=f"aliased_data__{node_alias}__any_lang_icontains=forty",
        )
        self.assertEqual(response.json()["count"], 1)
        self.assertEqual(
            response.json()["results"][0]["resourceinstanceid"],
            str(self.resource_42.pk),
        )

        response = self.client.get(
            reverse(
                "arches_querysets:api-tiles",
                kwargs={"graph": "datatype_lookups", "nodegroup_alias": "datatypes_1"},
            ),
            QUERY_STRING=f"aliased_data__{node_alias}__any_lang_icontains=forty",
        )
        self.assertEqual(response.json()["count"], 1)
        self.assertEqual(
            response.json()["results"][0]["resourceinstance"], str(self.resource_42.pk)
        )

        node_alias = "string_alias_n"
        response = self.client.get(
            reverse(
                "arches_querysets:api-tiles",
                kwargs={"graph": "datatype_lookups", "nodegroup_alias": "datatypes_n"},
            ),
            QUERY_STRING=f"aliased_data__{node_alias}__isnull=true",
        )
        self.assertEqual(response.json()["count"], 1)
        self.assertEqual(
            response.json()["results"][0]["resourceinstance"],
            str(self.resource_none.pk),
        )

    def test_filter_kwargs_disallows_joins(self):
        self.assertContains(
            self.client.get(
                reverse(
                    "arches_querysets:api-resources",
                    kwargs={"graph": "datatype_lookups"},
                ),
                QUERY_STRING=f"aliased_data__principaluser__username=dev",
            ),
            "Invalid filter param: principaluser",
            status_code=HTTPStatus.BAD_REQUEST,
        )

    def test_bogus_graph_slug(self):
        response = self.client.get(
            reverse("arches_querysets:api-resources", kwargs={"graph": "bogus"})
        )
        self.assertContains(
            response,
            "No nodes found for graph slug",
            status_code=HTTPStatus.BAD_REQUEST,
        )
        response = self.client.get(
            reverse(
                "arches_querysets:api-tiles",
                kwargs={"graph": "bogus", "nodegroup_alias": "bogus"},
            )
        )
        self.assertContains(
            response,
            "No nodes found for graph slug",
            status_code=HTTPStatus.BAD_REQUEST,
        )


class RestFrameworkPerformanceTests(GraphTestCase):
    @patch("arches_querysets.rest_framework.serializers.get_nodegroup_alias_lookup")
    def test_derivation_of_nodegroup_aliases(self, mocked_util):
        """Querying nodegroup aliases should only be done once in the view layer,
        not multiple times when building nested serializers. The serializer layer
        still has fallback code to support scripts, see test_bind_data_to_serializer(),
        but it shouldn't be called when using views.
        """
        self.client.get(
            reverse(
                "arches_querysets:api-resources", kwargs={"graph": "datatype_lookups"}
            )
        )
        mocked_util.assert_not_called()

    def test_resource_list_view_performance(self):
        # 1: auth
        # 2: auth groups
        # 3: node alias lookup in get_tiles()
        # 4-16: PerformanceTests.test_get_graph_objects()
        # 17: resource count (paginator)
        # 18: select resources limit 500
        # 19: tile depth 1
        # 20: resourcexresource depth 1
        # 21: tile depth 2
        # 22: resourcexresource depth 2
        # 23: tile depth 3: none!
        # 24: userprofile
        # 25-29: arches perms (BUG: core arches)
        with self.assertNumQueries(29):
            response = self.client.get(
                reverse(
                    "arches_querysets:api-resources",
                    kwargs={"graph": "datatype_lookups"},
                ),
                # Some datatypes are inefficient in fetching data for display values,
                # e.g. nodes, so make sure we're getting the resource with only Nones
                QUERY_STRING="aliased_data__number_alias__isnull=true",
            )
        self.assertContains(response, "datatypes_1_child", status_code=HTTPStatus.OK)
        self.assertEqual(response.json()["count"], 1)
        top_tile = response.json()["results"][0]["aliased_data"]["datatypes_1"]
        self.assertIsNone(top_tile["aliased_data"]["number_alias"]["node_value"])

    def test_tile_list_view_performance(self):
        # 1: auth
        # 2: auth groups
        # 3: node alias lookup in get_tiles()
        # 4-16: PerformanceTests.test_get_graph_objects()
        # 17: tile count (paginator)
        # 18: select tiles limit 500
        # 19: resourcexresource depth 1
        # 20: tile depth 2
        # 21: resourcexresource depth 2
        # 22: tile depth 3: none!
        # 23: userprofile
        # 24-28: arches perms (BUG: core arches)
        with self.assertNumQueries(28):
            response = self.client.get(
                reverse(
                    "arches_querysets:api-tiles",
                    kwargs={
                        "graph": "datatype_lookups",
                        "nodegroup_alias": "datatypes_1",
                    },
                ),
                # Some datatypes are inefficient in fetching data for display values,
                # e.g. nodes, so make sure we're getting the resource with only Nones
                QUERY_STRING="aliased_data__number_alias__isnull=true",
            )
        self.assertContains(response, "datatypes_1_child", status_code=HTTPStatus.OK)
        self.assertEqual(response.json()["count"], 1)
        top_tile = response.json()["results"][0]
        self.assertIsNone(top_tile["aliased_data"]["number_alias"]["node_value"])

    def test_resource_blank_view_performance(self):
        # 1-7: perms
        # 8: get_nodegroup_alias_lookup()
        # 9-12: NodeFetcherMixin._find_graph_nodes()
        with self.assertNumQueries(12):
            response = self.client.get(
                reverse(
                    "arches_querysets:api-resource-blank",
                    kwargs={"graph": "datatype_lookups"},
                )
            )
        self.assertContains(response, "datatypes_1_child")



import django.db.models

from arches.app.datatypes.datatypes import (
    BooleanDataType,
    DateDataType,
    DomainListDataType,
    NonLocalizedStringDataType,
    NumberDataType,
    GeojsonFeatureCollectionDataType,
)

from arches_querysets.datatypes import *
from arches_querysets.fields import (
    ConceptListField,
    DomainListField,
    LocalizedStringField,
    ResourceInstanceField,
    ResourceInstanceListField,
)


class DataTypeFactory(datatypes.DataTypeFactory):
    def get_instance(self, datatype):
        """Ensure every datatype has our additional methods."""
        instance = super().get_instance(datatype)

        if not hasattr(instance, "get_details"):
            instance.get_details = lambda value, *args, **kwargs: None
        if not hasattr(instance, "get_display_value_context_in_bulk"):
            instance.get_display_value_context_in_bulk = lambda *args, **kwargs: None
        if not hasattr(instance, "set_display_value_context_in_bulk"):
            instance.set_display_value_context_in_bulk = lambda *args, **kwargs: None

        return instance

    @staticmethod
    def get_model_field(instance):
        if model_field := getattr(instance, "model_field", None):
            return model_field
        match instance:
            case NumberDataType():
                return django.db.models.FloatField(null=True)
            case DateDataType():
                return django.db.models.DateTimeField(null=True)
            case BooleanDataType():
                return django.db.models.BooleanField(null=True)
            case NonLocalizedStringDataType():
                return django.db.models.TextField(null=True)
            case StringDataType():
                return LocalizedStringField(null=True)
            case ResourceInstanceListDataType():
                # must precede ResourceInstanceDataType
                return ResourceInstanceListField(null=True)
            case ResourceInstanceDataType():
                return ResourceInstanceField(null=True)
            case ConceptListDataType():
                return ConceptListField(null=True)
            case ConceptDataType(), NodeValueDataType():
                return django.db.models.UUIDField(null=True)
            case URLDataType():
                return django.db.models.JSONField(null=True)
            case DomainListDataType():
                return DomainListField(null=True)
            case FileListDataType():
                return django.db.models.JSONField(default=list, null=True)
            case GeojsonFeatureCollectionDataType():
                return django.db.models.JSONField(default=list, null=True)
            case _:
                return django.db.models.TextField(null=True)



from arches.app.datatypes import datatypes
from arches.app.models import models


# arches_version==9.0.0: remove
class NodeValueDataType(datatypes.NodeValueDataType):
    def get_display_value(self, tile, node, **kwargs):
        """Backport of Arches 8.1 version that moves value_node query under if gate."""
        from arches_querysets.datatypes.datatypes import DataTypeFactory

        datatype_factory = DataTypeFactory()
        try:
            data = self.get_tile_data(tile)
            tileid = data[str(node.nodeid)]
            if tileid:
                value_node = models.Node.objects.get(nodeid=node.config["nodeid"])
                value_tile = models.TileModel.objects.get(tileid=tileid)
                datatype = datatype_factory.get_instance(value_node.datatype)
                return datatype.get_display_value(value_tile, value_node)
            return ""
        except:
            raise Exception(
                f'Node with name "{node.name}" is not configured correctly.'
            )



from arches.app.datatypes import url


class URLDataType(url.URLDataType):
    def pre_structure_tile_data(self, tile, nodeid, **kwargs):
        if tile.data[nodeid] is None:
            tile.data[nodeid] = {"url": "", "url_label": ""}
            return
        return super().pre_structure_tile_data(tile, nodeid, **kwargs)



import uuid
from itertools import chain

from arches.app.datatypes import concept_types
from arches.app.models.models import Value
from arches.app.utils.betterJSONSerializer import JSONDeserializer, JSONSerializer


class ConceptDataType(concept_types.ConceptDataType):
    def transform_value_for_tile(self, value, **kwargs):
        if isinstance(value, dict) and (value_id := value.get("valueid")):
            return super().transform_value_for_tile(value_id, **kwargs)
        if isinstance(value, Value):
            return super().transform_value_for_tile(str(value.pk), **kwargs)
        return super().transform_value_for_tile(value, **kwargs)

    def to_python(self, value, **kwargs):
        return self.get_instance(value)

    def get_details(self, value, **kwargs):
        instance = self.get_instance(value)
        if not instance:
            return None
        return JSONDeserializer().deserialize(JSONSerializer().serialize([instance]))

    def get_instance(self, value):
        if value is None:
            return None
        try:
            value = uuid.UUID(value)
        except (TypeError, ValueError):
            pass
        return self.get_value(value)

    def get_display_value_context_in_bulk(self, values):
        value_ids = [
            value
            for value in values
            if value and uuid.UUID(value) not in self.value_lookup
        ]
        return Value.objects.filter(pk__in=value_ids)

    def set_display_value_context_in_bulk(self, datatype_context):
        for value in datatype_context:
            self.value_lookup[value.pk] = value


class ConceptListDataType(concept_types.ConceptListDataType):
    def transform_value_for_tile(self, value, **kwargs):
        if not value:
            return []
        if isinstance(value, list):
            if all(isinstance(val, Value) for val in value):
                return [str(val.pk) for val in value]
            return value
        return super().transform_value_for_tile(value, **kwargs)

    def to_python(self, value, **kwargs):
        return self.get_instances(value) or None

    def get_details(self, value, **kwargs):
        instances = self.get_instances(value)
        return JSONDeserializer().deserialize(JSONSerializer().serialize(instances))

    def get_instances(self, value):
        new_values = []
        for inner_value in value or []:
            try:
                new_val = self.get_value(uuid.UUID(inner_value))
            except (TypeError, ValueError):
                new_val = self.get_value(inner_value)
            new_values.append(new_val)
        return new_values

    def get_display_value(self, tile, node, **kwargs):
        new_values = []
        data = self.get_tile_data(tile)
        for val in data.get(str(node.nodeid)) or []:
            new_val = self.get_value(uuid.UUID(val))
            new_values.append(new_val.value)
        return ",".join(new_values)

    def get_display_value_context_in_bulk(self, values):
        not_null_values = [val for val in values if val]
        value_ids = list(chain(*not_null_values))
        return Value.objects.filter(pk__in=value_ids)

    def set_display_value_context_in_bulk(self, datatype_context):
        for value in datatype_context:
            self.value_lookup[value.pk] = value



from rest_framework import permissions

from arches.app.utils.permission_backend import group_required


class ReadOnly(permissions.BasePermission):
    def has_permission(self, request, view):
        return request.method in permissions.SAFE_METHODS


class Guest(permissions.BasePermission):
    def has_permission(self, request, view):
        return group_required(request.user, "Guest")


class RDMAdministrator(permissions.BasePermission):
    def has_permission(self, request, view):
        return group_required(request.user, "RDM Administrator")


class ResourceEditor(permissions.BasePermission):
    def has_permission(self, request, view):
        return group_required(request.user, "Resource Editor")



import uuid

from arches_querysets.models import ResourceTileTree, TileTree
from arches_querysets.utils.tests import GraphTestCase


class LookupTestCase(GraphTestCase):
    def setUp(self):
        self.resources = ResourceTileTree.get_tiles("datatype_lookups")
        self.tiles_1 = TileTree.get_tiles(
            "datatype_lookups", nodegroup_alias="datatypes_1"
        )
        self.tiles_n = TileTree.get_tiles(
            "datatype_lookups", nodegroup_alias="datatypes_n"
        )


class GenericLookupTests(LookupTestCase):
    def test_cardinality_1(self):
        # Exact
        for lookup, value in [
            ("boolean_alias", True),
            ("number_alias", 42.0),  # Use a float so that stringifying causes failure.
            ("url_alias__url_label", "42.com"),
            ("non_localized_string_alias", "forty-two"),
            ("string_alias__en__value", "forty-two"),
            ("date_alias", "2042-04-02"),
            # More natural lookups in ResourceInstanceLookupTests
            ("resource_instance_alias__0__resourceId", str(self.resource_42.pk)),
            ("resource_instance_list_alias__0__resourceId", str(self.resource_42.pk)),
            ("concept_alias", self.concept_value.pk),
            ("concept_list_alias", [str(self.concept_value.pk)]),
            ("node_value_alias", self.cardinality_1_tile.pk),
        ]:
            with self.subTest(lookup=lookup, value=value):
                self.assertTrue(self.resources.filter(**{lookup: value}))
                self.assertTrue(self.tiles_1.filter(**{lookup: value}))

    def test_cardinality_n(self):
        # Contains
        for lookup, value in [
            ("boolean_alias_n__contains", [True]),
            ("number_alias_n__contains", [42.0]),
            ("url_alias_n__0__url_label", "42.com"),
            ("non_localized_string_alias_n__contains", ["forty-two"]),
            ("date_alias_n__contains", ["2042-04-02"]),
            # better lookups for RI{list} below.
            ("resource_instance_alias_n__0__0__resourceId", str(self.resource_42.pk)),
            # you likely want ids_contain, below.
            (
                "resource_instance_list_alias_n__0__0__resourceId",
                str(self.resource_42.pk),
            ),
            ("concept_alias_n__contains", [self.concept_value.pk]),
            # you likely want any_contains, below.
            ("concept_list_alias_n__0__contains", str(self.concept_value.pk)),
            ("node_value_alias_n__contains", [self.cardinality_n_tile.pk]),
        ]:
            with self.subTest(lookup=lookup, value=value):
                self.assertTrue(self.resources.filter(**{lookup: value}))

    def test_cardinality_1_values(self):
        for node in self.data_nodes_1:
            with self.subTest(alias=node.alias):
                self.assertTrue(self.resources.values(node.alias))
                self.assertTrue(self.tiles_1.values(node.alias))
                self.assertTrue(self.resources.values_list(node.alias))
                self.assertTrue(self.tiles_1.values_list(node.alias))

    def test_cardinality_n_values(self):
        for node in self.data_nodes_n:
            with self.subTest(alias=node.alias):
                self.assertTrue(self.resources.values(node.alias))
                self.assertTrue(self.tiles_n.values(node.alias))
                self.assertTrue(self.resources.values_list(node.alias))
                self.assertTrue(self.tiles_n.values_list(node.alias))

    def test_values_path_transforms(self):
        resources = self.resources.exclude(resource_instance_list_alias=None)
        values = resources.values("resource_instance_list_alias__0__resourceId")
        uuid_val = values[0]["resource_instance_list_alias__0__resourceId"]
        # Implicitly test the result is a uuid
        uuid.UUID(uuid_val)

        values = resources.values_list(
            "resource_instance_list_alias__0__resourceId", flat=True
        )
        uuid_val = values[0]
        uuid.UUID(uuid_val)


class NonLocalizedStringLookupTests(LookupTestCase):
    def test_cardinality_1(self):
        self.assertTrue(
            self.resources.filter(non_localized_string_alias__contains="forty")
        )

    def test_cardinality_n(self):
        self.assertTrue(
            self.resources.filter(non_localized_string_alias_n__contains=["forty-two"])
        )
        self.assertFalse(
            self.resources.filter(non_localized_string_alias_n__contains=["forty"])
        )
        self.assertTrue(
            self.resources.filter(non_localized_string_alias_n__any_contains="forty")
        )
        self.assertFalse(
            self.resources.filter(non_localized_string_alias_n__any_contains="FORTY")
        )
        self.assertTrue(
            self.resources.filter(non_localized_string_alias_n__any_icontains="FORTY")
        )


class LocalizedStringLookupTests(LookupTestCase):
    def test_cardinality_1(self):
        for lookup, value in [
            ("string_alias__any_lang_startswith", "forty"),
            ("string_alias__any_lang_istartswith", "FORTY"),
            ("string_alias__any_lang_contains", "fort"),
            ("string_alias__any_lang_icontains", "FORT"),
        ]:
            with self.subTest(lookup=lookup, value=value):
                self.assertTrue(self.resources.filter(**{lookup: value}))

        # Negatives
        for lookup, value in [
            ("string_alias__any_lang_startswith", "orty-two"),
            ("string_alias__any_lang_istartswith", "ORTY-TWO"),
            ("string_alias__any_lang_contains", "orty-three"),
            ("string_alias__any_lang_icontains", "ORTY-THREE"),
        ]:
            with self.subTest(lookup=lookup, value=value):
                self.assertFalse(self.resources.filter(**{lookup: value}))

    def test_cardinality_n(self):
        for lookup, value in [
            ("string_alias_n__any_lang_startswith", "forty"),
            ("string_alias_n__any_lang_istartswith", "FORTY"),
            ("string_alias_n__any_lang_contains", "fort"),
            ("string_alias_n__any_lang_icontains", "FORT"),
        ]:
            with self.subTest(lookup=lookup, value=value):
                self.assertTrue(self.resources.filter(**{lookup: value}))

        # Negatives
        for lookup, value in [
            ("string_alias_n__any_lang_startswith", "orty-two"),
            ("string_alias_n__any_lang_istartswith", "ORTY-TWO"),
            ("string_alias_n__any_lang_contains", "orty-three"),
            ("string_alias_n__any_lang_icontains", "ORTY-THREE"),
        ]:
            with self.subTest(lookup=lookup, value=value):
                self.assertFalse(self.resources.filter(**{lookup: value}))


class ResourceInstanceLookupTests(LookupTestCase):
    def test_cardinality_1(self):
        for lookup, value in [
            ("resource_instance_alias__id", str(self.resource_42.pk)),
            ("resource_instance_list_alias__contains", str(self.resource_42.pk)),
        ]:
            with self.subTest(lookup=lookup, value=value):
                self.assertTrue(self.resources.filter(**{lookup: value}))

    def test_cardinality_n(self):
        for lookup, value in [
            ("resource_instance_alias_n__ids_contain", str(self.resource_42.pk)),
            ("resource_instance_list_alias_n__ids_contain", str(self.resource_42.pk)),
        ]:
            with self.subTest(lookup=lookup, value=value):
                self.assertTrue(self.resources.filter(**{lookup: value}))



import json
from unittest.mock import Mock

from arches_querysets.models import ResourceTileTree
from arches_querysets.utils.tests import GraphTestCase


class DatatypeRepresentationTests(GraphTestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        resources = ResourceTileTree.get_tiles(
            "datatype_lookups", as_representation=True
        )
        cls.resource_42 = resources.get(pk=cls.resource_42.pk)
        cls.resource_none = resources.get(pk=cls.resource_none.pk)
        cls.datatype_1 = cls.resource_42.aliased_data.datatypes_1
        cls.datatype_n = cls.resource_42.aliased_data.datatypes_n
        cls.datatype_1_none = cls.resource_none.aliased_data.datatypes_1
        cls.datatype_n_none = cls.resource_none.aliased_data.datatypes_n

    def test_as_representation_display_values(self):
        display_values = {
            # Start with the tile representation.
            **self.sample_data_1,
            # Some representations are different.
            # Boolean resolves to a string.
            "boolean": str(self.sample_data_1["boolean"]),
            # Number resolves to a string. TODO: localize?
            "number": str(self.sample_data_1["number"]),
            # String resolves to active language.
            "string": "forty-two",
            # Resource Instance{list} resolves to localized name.
            "resource-instance": "Resource referencing 42",
            "resource-instance-list": "Resource referencing 42",
            # Concept{list} resolves to concept value.
            "concept": "Arches",
            "concept-list": "Arches",
            # Node value resolves to node value.
            "node-value": self.sample_data_1["date"],
            # URL resolves to the stringified dictionary.
            "url": json.dumps(self.sample_data_1["url"]),
            # File-list resolves to urls joined with " | ".
            "file-list": self.sample_data_1["file-list"][0]["url"],
        }

        # The representation is available on the nodegroup .aliased_data.
        for datatype, representation in display_values.items():
            node_alias = datatype.replace("-", "_") + "_alias"
            for resource_data, resource_data_none, cardinality in [
                (self.datatype_1.aliased_data, self.datatype_1_none.aliased_data, "1"),
                (
                    self.datatype_n[0].aliased_data,
                    self.datatype_n_none[0].aliased_data,
                    "n",
                ),
            ]:
                with self.subTest(datatype=datatype, cardinality=cardinality):
                    lookup = node_alias if cardinality == "1" else node_alias + "_n"
                    value = getattr(resource_data, lookup)
                    self.assertEqual(value.get("display_value"), representation)
                    value = getattr(resource_data_none, lookup)
                    self.assertEqual(value.get("display_value"), "")

    def test_as_representation_details(self):
        detail_values = {
            # Resource Instance & list details both resolve to an array of objects.
            "resource-instance": [
                {
                    "resource_id": str(self.resource_42.pk),
                    "display_value": self.resource_42.descriptors["en"]["name"],
                    # This can be extended as necessary via get_details()
                }
            ],
            "resource-instance-list": [
                {
                    "resource_id": str(self.resource_42.pk),
                    "display_value": self.resource_42.descriptors["en"]["name"],
                }
            ],
            # Same with concept & concept-list.
            "concept": [
                {
                    "valueid": "d8c60bf4-e786-11e6-905a-b756ec83dad5",
                    "concept_id": "00000000-0000-0000-0000-000000000001",
                    "valuetype_id": "prefLabel",
                    "value": "Arches",
                    "language_id": "en",
                }
            ],
            "concept-list": [
                {
                    "valueid": "d8c60bf4-e786-11e6-905a-b756ec83dad5",
                    "concept_id": "00000000-0000-0000-0000-000000000001",
                    "valuetype_id": "prefLabel",
                    "value": "Arches",
                    "language_id": "en",
                }
            ],
        }

        # The details are available on the nodegroup .aliased_data.
        for datatype, details in detail_values.items():
            node_alias = datatype.replace("-", "_") + "_alias"
            for resource_data, resource_data_none, cardinality in [
                (self.datatype_1.aliased_data, self.datatype_1_none.aliased_data, "1"),
                (
                    self.datatype_n[0].aliased_data,
                    self.datatype_n_none[0].aliased_data,
                    "n",
                ),
            ]:
                lookup = node_alias if cardinality == "1" else node_alias + "_n"
                if lookup == "node_value_alias_n":
                    details = self.sample_data_n["node-value"]
                with self.subTest(lookup=lookup):
                    value = getattr(resource_data, lookup)
                    self.assertEqual(value.get("details"), details)
                    value = getattr(resource_data_none, lookup)
                    if value.get("details") is not None:
                        self.assertEqual(value.get("details"), [])


class DatatypePythonTests(GraphTestCase):
    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        resources = ResourceTileTree.get_tiles(
            "datatype_lookups", as_representation=False
        )
        cls.resource = resources.get(pk=cls.resource_42.pk)
        cls.resource_none = resources.get(pk=cls.resource_none.pk)
        cls.datatype_1 = cls.resource.aliased_data.datatypes_1
        cls.datatype_n = cls.resource.aliased_data.datatypes_n
        cls.datatype_1_none = cls.resource_none.aliased_data.datatypes_1
        cls.datatype_n_none = cls.resource_none.aliased_data.datatypes_n

    def test_python_values(self):
        python_values = {
            # Start with the tile representation.
            **self.sample_data_1,
            # Some python values are different.
            # Resource Instances become model instances
            "resource-instance": self.resource_42,
            "resource-instance-list": [self.resource_42],
            # Concepts become concept value model instances
            "concept": self.concept_value,
            "concept-list": [self.concept_value],
        }

        # The python value is available on the nodegroup .aliased_data.
        for datatype, python_value in python_values.items():
            node_alias = datatype.replace("-", "_") + "_alias"
            for resource_data, resource_data_none, cardinality in [
                (self.datatype_1.aliased_data, self.datatype_1_none.aliased_data, "1"),
                (
                    self.datatype_n[0].aliased_data,
                    self.datatype_n_none[0].aliased_data,
                    "n",
                ),
            ]:
                lookup = node_alias if cardinality == "1" else node_alias + "_n"
                if lookup == "node_value_alias_n":
                    python_value = self.sample_data_n["node-value"]
                with self.subTest(lookup=lookup):
                    value = getattr(resource_data, lookup)
                    self.assertEqual(value, python_value)
                    value = getattr(resource_data_none, lookup)
                    self.assertIsNone(value, datatype)


class DatatypeMethodTests(GraphTestCase):
    def test_transform_value_for_tile(self):
        test_values = {
            # TODO - add more datatypes tests here.
            "file-list": [
                {
                    "input": "commma-separated-list.png,from-bulk-data-manager.jpg",
                    "output": [
                        {
                            "status": "uploaded",
                            "name": "commma-separated-list.png",
                            "type": "image/png",
                            "file_id": "b503428b-1e04-4c79-8188-f70837461a07",
                            "url": "/files/b503428b-1e04-4c79-8188-f70837461a07",
                            "accepted": True,
                            "renderer": "5e05aa2e-5db0-4922-8938-b4d2b7919733",
                        },
                        {
                            "status": "uploaded",
                            "name": "from-bulk-data-manager.jpg",
                            "type": "image/jpeg",
                            "file_id": "35a11db0-3796-4b25-becd-b369096a0942",
                            "url": "/files/35a11db0-3796-4b25-becd-b369096a0942",
                            "accepted": True,
                            "renderer": "5e05aa2e-5db0-4922-8938-b4d2b7919733",
                        },
                    ],
                    "equality_test": lambda list1, list2: list1[0].keys()
                    == list2[0].keys(),
                },
                {
                    "input": [
                        {
                            "status": "uploaded",
                            "name": "commma-separated-list.png",
                            "type": "image/png",
                            "file_id": "b503428b-1e04-4c79-8188-f70837461a07",
                            "url": "/files/b503428b-1e04-4c79-8188-f70837461a07",
                            "accepted": True,
                            "renderer": "5e05aa2e-5db0-4922-8938-b4d2b7919733",
                            "altText": {
                                "en": {
                                    "value": "Replaceable alt text",
                                    "direction": "ltr",
                                }
                            },
                            "attribution": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                            "description": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                            "title": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                        },
                        {
                            "status": "uploaded",
                            "name": "from-bulk-data-manager.jpg",
                            "type": "image/jpeg",
                            "file_id": "35a11db0-3796-4b25-becd-b369096a0942",
                            "url": "/files/35a11db0-3796-4b25-becd-b369096a0942",
                            "accepted": True,
                            "renderer": "5e05aa2e-5db0-4922-8938-b4d2b7919733",
                            "altText": {
                                "en": {
                                    "value": "Replaceable alt text",
                                    "direction": "ltr",
                                }
                            },
                            "attribution": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                            "description": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                            "title": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                        },
                    ],
                    "output": [
                        {
                            "status": "uploaded",
                            "name": "commma-separated-list.png",
                            "type": "image/png",
                            "file_id": "b503428b-1e04-4c79-8188-f70837461a07",
                            "url": "/files/b503428b-1e04-4c79-8188-f70837461a07",
                            "accepted": True,
                            "renderer": "5e05aa2e-5db0-4922-8938-b4d2b7919733",
                            "altText": {
                                "en": {
                                    "value": "Replaceable alt text",
                                    "direction": "ltr",
                                }
                            },
                            "attribution": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                            "description": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                            "title": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                        },
                        {
                            "status": "uploaded",
                            "name": "from-bulk-data-manager.jpg",
                            "type": "image/jpeg",
                            "file_id": "35a11db0-3796-4b25-becd-b369096a0942",
                            "url": "/files/35a11db0-3796-4b25-becd-b369096a0942",
                            "accepted": True,
                            "renderer": "5e05aa2e-5db0-4922-8938-b4d2b7919733",
                            "altText": {
                                "en": {
                                    "value": "Replaceable alt text",
                                    "direction": "ltr",
                                }
                            },
                            "attribution": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                            "description": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                            "title": {
                                "en": {
                                    "value": "",
                                    "direction": "ltr",
                                }
                            },
                        },
                    ],
                    "equality_test": lambda list1, list2: list1[0].keys()
                    == list2[0].keys(),
                },
            ],
            "resource-instance": [
                {
                    "input": self.resource_42.pk,  # test as uuid
                    "output": [
                        {
                            "resourceId": str(self.resource_42.pk),
                            "ontologyProperty": "",
                            "inverseOntologyProperty": "",
                        }
                    ],
                },
                {
                    "input": str(self.resource_42.pk),  # test as string
                    "output": [
                        {
                            "resourceId": str(self.resource_42.pk),
                            "ontologyProperty": "",
                            "inverseOntologyProperty": "",
                        }
                    ],
                },
                {
                    "input": self.resource_42,  # test as model instance
                    "output": [
                        {
                            "resourceId": str(self.resource_42.pk),
                            "ontologyProperty": "",
                            "inverseOntologyProperty": "",
                        }
                    ],
                },
                {
                    "input": json.dumps(
                        [
                            {  # test as stringified dict
                                "resourceId": str(self.resource_42.pk),
                                "ontologyProperty": "testProperty",
                                "inverseOntologyProperty": "testInverseProperty",
                            }
                        ]
                    ),
                    "output": [
                        {
                            "resourceId": str(self.resource_42.pk),
                            "ontologyProperty": "testProperty",
                            "inverseOntologyProperty": "testInverseProperty",
                        }
                    ],
                },
                {
                    "input": [
                        {  # test as dict
                            "resourceId": str(self.resource_42.pk),
                            "ontologyProperty": "testProperty",
                            "inverseOntologyProperty": "testInverseProperty",
                        }
                    ],
                    "output": [
                        {
                            "resourceId": str(self.resource_42.pk),
                            "ontologyProperty": "testProperty",
                            "inverseOntologyProperty": "testInverseProperty",
                        }
                    ],
                },
            ],
            "resource-instance-list": [
                {
                    "input": [self.resource_42.pk],
                    "output": [
                        {
                            "resourceId": str(self.resource_42.pk),
                            "ontologyProperty": "",
                            "inverseOntologyProperty": "",
                        }
                    ],
                }
            ],
        }

        for datatype, cases in test_values.items():
            for value in cases:
                with self.subTest(datatype=datatype, value=value["input"]):
                    datatype_instance = self.datatype_factory.get_instance(
                        datatype=datatype
                    )
                    transformed_value = datatype_instance.transform_value_for_tile(
                        value["input"], is_existing_tile=False
                    )
                    if equality_test := value.get("equality_test"):
                        self.assertTrue(
                            equality_test(transformed_value, value["output"])
                        )
                    else:
                        self.assertEqual(transformed_value, value["output"])



# Generated by Django 5.2b1 on 2025-03-04 13:37

from django.db import migrations


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("models", "11499_add_editlog_resourceinstance_idx"),  # last 7.6 migration
    ]

    operations = [
        migrations.CreateModel(
            name="GraphWithPrefetching",
            fields=[],
            options={
                "db_table": "graphs",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("models.graphmodel",),
        ),
        migrations.CreateModel(
            name="ResourceTileTree",
            fields=[],
            options={
                "db_table": "resource_instances",
                "permissions": (("no_access_to_resourceinstance", "No Access"),),
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("models.resourceinstance",),
        ),
        migrations.CreateModel(
            name="TileTree",
            fields=[],
            options={
                "db_table": "tiles",
                "proxy": True,
                "indexes": [],
                "constraints": [],
            },
            bases=("models.tilemodel",),
        ),
    ]



from arches_querysets.datatypes import DataTypeFactory
from arches_querysets.models import GraphWithPrefetching, ResourceTileTree, TileTree
from arches_querysets.utils.tests import GraphTestCase


class PerformanceTests(GraphTestCase):
    def test_get_graph_objects(self):
        # 1: graph
        # 2: graph -> node
        # 3: graph -> node -> cardxnodexwidget
        # 4: graph -> node -> nodegroup
        # 5: graph -> node -> nodegroup -> node
        # 6: graph -> node -> nodegroup -> node -> cardxnodexwidget
        # 7: graph -> node -> nodegroup -> card
        # (8-12): 3-7 again for (depth 2 nodegroup)
        # 13: depth 3 nodegroup (none!)
        with self.assertNumQueries(13):
            GraphWithPrefetching.objects.prefetch("datatype_lookups").get()

    def test_get_resources(self):
        # Clear the value lookups to avoid flakiness.
        factory = DataTypeFactory()
        concept_dt = factory.get_instance("concept")
        concept_dt.value_lookup = {}
        concept_list_dt = factory.get_instance("concept-list")
        concept_list_dt.value_lookup = {}

        # 1: nodes
        # 2: resources
        # 3-15: test_get_graph_objects()
        # 16: tile depth 1
        # 17: resource -> resourcexresource depth 1
        # 18: resource -> resourcexresource -> to_resource
        # (17-18 are a little unfortunate, but worth it for resourcexresource prefetches.
        # 19-21: depth 2
        # 22: concept value
        # 23-24: (N+1 BUG: core arches) more concept values
        with self.assertNumQueries(24):
            qs = ResourceTileTree.get_tiles("datatype_lookups")
            self.assertCountEqual(qs, [self.resource_42, self.resource_none])

    def test_get_tiles(self):
        # 1-22 from test_get_resources()
        with self.assertNumQueries(22):
            qs = TileTree.get_tiles("datatype_lookups", "datatypes_1")
            self.assertCountEqual(
                qs, [self.cardinality_1_tile, self.cardinality_1_tile_none]
            )



from django.db.models import Max, Min

from arches_querysets.models import ResourceTileTree
from arches_querysets.utils.tests import GraphTestCase


class AggregateTests(GraphTestCase):
    def test_number(self):
        self.resources = ResourceTileTree.get_tiles("datatype_lookups")

        # Edit the resource that usually has None in all nodes to have a value of 43.
        # Fetch it since cls.resource_none was not fetched via get_tiles() :/
        resource2 = self.resources.get(pk=self.resource_none.pk)
        resource2.aliased_data.datatypes_1.aliased_data.number_alias = 43
        resource2.aliased_data.datatypes_n[0].aliased_data.number_alias_n = 43
        resource2.save(force_admin=True)

        # Per-table aggregate on cardinality-1 value
        query = self.resources.aggregate(Min("number_alias"), Max("number_alias"))
        self.assertEqual(query["number_alias__min"], 42.0)
        self.assertEqual(query["number_alias__max"], 43.0)

        # Per-table aggregate on arrays, e.g. [43] > [42], but [43, 42] < [43, 44]
        query = self.resources.aggregate(Min("number_alias_n"), Max("number_alias_n"))
        self.assertEqual(query["number_alias_n__min"], [42.0])
        self.assertEqual(query["number_alias_n__max"], [43.0])



import logging
from arches.app.models.resource import Resource
from celery import shared_task


@shared_task
def index_resource(resource_id):
    logger = logging.getLogger(__name__)
    try:
        resource = Resource.objects.filter(pk=resource_id).get()
        resource.index()
    except Resource.DoesNotExist:
        logger.error(f"Resource with id {resource_id} does not exist.")
    except Exception as e:
        logger.error(f"Error indexing resource {resource_id}: {e}")



from arches_querysets.models import TileTree
from arches_querysets.utils.tests import GraphTestCase


class InternalsTests(GraphTestCase):
    def test_node_alias_collision_with_model_field(self):
        self.file_list_node_1.alias = "file"
        self.file_list_node_1.save()
        # Previously, this clashed with the related query name "file"
        qs = TileTree.get_tiles(
            graph_slug="datatype_lookups", nodegroup_alias="datatypes_1"
        )
        # TODO: determine the reserved namespace to use here.
        self.assertEqual(
            qs.filter(_arches_querysets_file__isnull=True)[0].resourceinstance_id,
            self.resource_none.pk,
        )



from django.db.models.lookups import Lookup, PatternLookup, Transform
from psycopg2.extensions import AsIs, QuotedString

from arches_querysets.fields import (
    CardinalityNResourceInstanceField,
    CardinalityNResourceInstanceListField,
    CardinalityNLocalizedStringField,
    CardinalityNTextField,
    ResourceInstanceListField,
    ResourceInstanceField,
    LocalizedStringField,
)


class JSONPathFilter:
    def process_rhs(self, compiler, connection):
        rhs, params = super().process_rhs(compiler, connection)
        if '"' in params[0]:
            raise ValueError("Double quotes are not allowed in JSONPath filters.")
        quoted = AsIs(QuotedString(params[0]).getquoted().decode()[1:-1])
        return rhs, (quoted,)


@CardinalityNTextField.register_lookup
class AnyContains(PatternLookup):
    """Provide a single string. Adapted from https://code.djangoproject.com/ticket/34942"""

    lookup_name = "any_contains"
    like_operator = "LIKE"

    def as_sql(self, compiler, connection):
        # Avoid connection.ops.lookup_cast in BuiltinLookup.process_lhs()
        lhs, lhs_params = Lookup.process_lhs(self, compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return (
            "EXISTS(SELECT * FROM UNNEST(%s) AS a WHERE a %s %s)"
            % (lhs, self.like_operator, rhs),
            params,
        )


@CardinalityNTextField.register_lookup
class AnyIContains(AnyContains):
    lookup_name = "any_icontains"
    like_operator = "ILIKE"


@LocalizedStringField.register_lookup
class AnyLanguageStartsWith(JSONPathFilter, Lookup):
    lookup_name = "any_lang_startswith"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return "%s @? '$.*.value ? (@ starts with \"%s\")'" % (lhs, rhs), params


@LocalizedStringField.register_lookup
class AnyLanguageIStartsWith(JSONPathFilter, Lookup):
    lookup_name = "any_lang_istartswith"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return (
            '%s @? \'$.*.value ? (@ like_regex "^%s" flag "i")\'' % (lhs, rhs),
            params,
        )


@LocalizedStringField.register_lookup
class AnyLanguageContains(JSONPathFilter, Lookup):
    lookup_name = "any_lang_contains"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return "%s @? '$.*.value ? (@ like_regex \"%s\")'" % (lhs, rhs), params


@LocalizedStringField.register_lookup
class AnyLanguageIContains(JSONPathFilter, Lookup):
    lookup_name = "any_lang_icontains"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return '%s @? \'$.*.value ? (@ like_regex "%s" flag "i")\'' % (lhs, rhs), params


@ResourceInstanceField.register_lookup
class ResourceInstanceId(Transform):
    lookup_name = "id"
    template = "(%(expressions)s -> 0 -> 'resourceId')"


@CardinalityNResourceInstanceField.register_lookup
class ArrayResourceInstanceId(JSONPathFilter, Lookup):
    lookup_name = "ids_contain"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return (
            "TO_JSONB(%s) @? '$[*][*].resourceId ? (@ == \"%s\")'" % (lhs, rhs),
            params,
        )


@CardinalityNResourceInstanceListField.register_lookup
class ArrayAnyResourceInstanceId(JSONPathFilter, Lookup):
    lookup_name = "ids_contain"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return (
            "TO_JSONB(%s) @? '$[*][*][*].resourceId ? (@ == \"%s\")'" % (lhs, rhs),
            params,
        )


@ResourceInstanceListField.register_lookup
class ResourceInstanceListContains(JSONPathFilter, Lookup):
    lookup_name = "contains"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return "%s @? '$[*].resourceId ? (@ == \"%s\")'" % (lhs, rhs), params


@CardinalityNLocalizedStringField.register_lookup
class ArrayAnyLanguageEquals(JSONPathFilter, Lookup):
    lookup_name = "any_lang"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return "TO_JSONB(%s) @? '$[*].*.value ? (@ == \"%s\")'" % (lhs, rhs), params


@CardinalityNLocalizedStringField.register_lookup
class ArrayAnyLanguageContains(JSONPathFilter, Lookup):
    lookup_name = "any_lang_contains"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return (
            "TO_JSONB(%s) @? '$[*].*.value ? (@ like_regex \"%s\")'" % (lhs, rhs),
            params,
        )


@CardinalityNLocalizedStringField.register_lookup
class ArrayAnyLanguageIContains(JSONPathFilter, Lookup):
    lookup_name = "any_lang_icontains"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return (
            'TO_JSONB(%s) @? \'$[*].*.value ? (@ like_regex "%s" flag "i")\''
            % (lhs, rhs),
            params,
        )


@CardinalityNLocalizedStringField.register_lookup
class ArrayAnyLanguageStartsWith(JSONPathFilter, Lookup):
    lookup_name = "any_lang_startswith"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return (
            "TO_JSONB(%s) @? '$[*].*.value ? (@ starts with \"%s\")'" % (lhs, rhs),
            params,
        )


@CardinalityNLocalizedStringField.register_lookup
class ArrayAnyLanguageIStartsWith(JSONPathFilter, Lookup):
    lookup_name = "any_lang_istartswith"

    def as_sql(self, compiler, connection):
        lhs, lhs_params = self.process_lhs(compiler, connection)
        rhs, rhs_params = self.process_rhs(compiler, connection)
        params = (*lhs_params, *rhs_params)
        return (
            'TO_JSONB(%s) @? \'$[*].*.value ? (@ like_regex "^%s" flag "i")\''
            % (lhs, rhs),
            params,
        )



import json

from rest_framework.parsers import MultiPartParser, DataAndFiles
from rest_framework.exceptions import ParseError


class MultiPartJSONParser(MultiPartParser):
    """
    Parses multipart/form-data, extracts the 'json' part,
    decodes it into a native dict, and returns that as request.data.
    Uploaded files remain in request.FILES.
    """

    def parse(self, stream, media_type=None, parser_context=None):
        parsed = super().parse(stream, media_type, parser_context)

        data_fields = parsed.data
        file_fields = parsed.files

        raw_json = data_fields.get("json") or file_fields.get("json").read()
        if raw_json is None:
            raise ParseError(detail="Missing 'json' part in multipart payload")

        try:
            decoded = json.loads(raw_json)
        except ValueError as err:
            raise ParseError(detail=f"Invalid JSON in 'json' part: {err}")

        if not isinstance(decoded, dict):
            raise ParseError(detail="'json' part did not decode to a dict")

        return DataAndFiles(decoded, file_fields)



from django.core.management.base import BaseCommand
from django.db import transaction

from arches.app.models import models


class Command(BaseCommand):
    def handle(self, *args, **options):
        self.stdout.write("Deleting test data...")
        with transaction.atomic():
            models.ResourceInstance.objects.filter(
                graph__slug="datatype_lookups"
            ).delete()
            models.GraphModel.objects.filter(slug="datatype_lookups").delete()
        self.stdout.write("Finished!")
